# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module simulates Alice's operations"""

import logging
import sys

from base64 import urlsafe_b64encode

from QKDSimkit.core.qexceptions import qsocketerror, aliceerror
from QKDSimkit.core.sender import Sender


def import_key(channel_address: str, ID: str, size: int = 256):
    """Alice's procedure to agree on a shared key

    Args:
        channel_address (str): channel address
        ID (str): identifier of alice-bob pair
        size (int): size of key in bits
    """
    channelIP, channelPort = channel_address.split(':')
    channelPort = int(channelPort)

    for count in range(0, 1000):
        alice = Sender(ID, size)
        try:
            # connect to quantum channel
            alice.connect_to_channel(channelIP, channelPort)
            # create and send a photon pulse through the quantum channel
            photon_pulse = alice.create_photon_pulse()
            alice.send_photon_pulse(photon_pulse)
        except qsocketerror as err:
            raise aliceerror("Connection error while connecting to the quantum channel (" + str(err) + "). Disconnecting.")
        except Exception as e:
            raise aliceerror('Generic error during qunatum photon exchange: ' + str(e))

        # connect to classic channel
        try:
            alice.reset_socket()
            alice.connect_to_channel(channelIP, channelPort)
        except qsocketerror as err:
            raise aliceerror("Connection error while connecting to the classic channel (" + str(err) + "). Disconnecting.")

        # exchange basis
        try:
            # listen for Bob's basis
            alice.listen_for('bob', 'other_bases')
            # generate the reconciled key from Bob's basis
            alice.generate_reconciled_key()
            # create and send the basis through the classic channel
            alice.send('alice-reconciled_key', repr(alice.reconciled_key))
        except qsocketerror as err:
            raise aliceerror("Connection error while exchanging bases (" + str(err) + "). Disconnecting.")
        except Exception as err:
            raise aliceerror("Generic error while exchanging bases: " + str(err))

        # create key and public sub key
        alice.create_keys()

        # exchange sub key
        try:
            # send the public sub key
            alice.send('alice-other_sub_key', repr(alice.sub_shared_key))
            # listen for Bob's public sub key
            alice.listen_for('bob', 'other_sub_key')

            alice.decision = alice.validate()

            # send decision
            alice.send('alice-other_decision', repr(alice.decision))
            # listen for Alice's sub key
            alice.listen_for('bob', 'other_decision')
        except qsocketerror as err:
            raise aliceerror("Connection error while comparing sub_keys (" + str(err) + "). Disconnecting.")
        except Exception as err:
            raise aliceerror("Generic error while comparing sub_keys (" + str(err) + "). Disconnecting.")


        alice.reset_socket()

        # choose what to do
        if alice.decision == alice.other_decision and alice.decision == 1:
            # return a correct key
            alice.get_key()
            logging.info("Success!")
            alice.key = alice.key[:size]
            alice.key = [int("".join(map(str, alice.key[i:i + 8])), 2) for i in range(0, len(alice.key), 8)]
            return urlsafe_b64encode(bytearray(alice.key))
        elif alice.decision == alice.other_decision and alice.decision == 0:
            # retry
            logging.warning("Failed to match key, trying again")
            continue
        else:
            # exit
            raise aliceerror("Failed! Noise or eavesdropper detected")
    raise aliceerror("Error: too many attempts to find a shared key")


if __name__ == '__main__':
    try:
        import_key('127.0.0.1:8000', 'id', 256)
    except aliceerror as ae:
        logging.error(str(ae))



# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

import logging
import socket
import sys

from .models import Photon
from .node import Node
from .qexceptions import qsocketerror

logging.basicConfig(level=logging.DEBUG)


class Receiver(Node):
    """Receiver class, it expands node, it contains methods to communicate a sender node, it can't take action but it
    has to wait for the sender node for sending data, it can answer to a request with a message or to a message with an
    acknowledgement"""
    def __init__(self, ID, size):
        super().__init__(ID, size)
        self.polarization_vector = []
        self.sent_acks = []
        self.sent_messages = {}

    def measure_photon_pulse(self):
        """Measure photon pulse

        given the vector that stores polarizations of received photons it creates a list of photons each polarization,
        basis and bit will be determined byt the measure method according to physical properties, a list with the basis
        of every photon will be stored
        """
        for p in range(len(self.polarization_vector)):
            self.photon_pulse.append(Photon())
            self.photon_pulse[p].polarization = self.photon_pulse[p].measure(int(self.polarization_vector[p]))
            self.photon_pulse[p].bit = self.photon_pulse[p].set_bit_from_measurement()
        self.bases = [p.basis for p in self.photon_pulse]

    def listen_quantum(self):
        """ Listen method to receive photon pulse
        it behaves like a wrapper for recv for photon pulses
        """
        try:
            logging.info("listening to quantum channel for photon pulse...")
            while True:
                message = self.recv('qpulse')
                self.polarization_vector = message.split("~")[:-1]
                break
        except socket.error:
            raise qsocketerror("not connected to any channel")

    def recv(self, header: str) -> str:
        """Receive function for receiver node
        it listen for a message, in case the header of the received message doesn't match it checks if an acknowledgment
        for the received message has been already sent or if the received message is an acknowledgement itself for a
        previously sent message, it sends a new acknowledgment in the first case and it sends again the message in the
        other case, if the message is new and the header is correct it sends an acknowledgement and it saves it
        Args:
            header (str): unique identifier of the message that has to be received
        Returns:
            message (str): the payload of the received data, the header and some other infos are not returned
        """
        try:
            for i in range(self.connection_attempts):  # loop for different messages
                received = self.recv_all()
                if not received:
                    continue
                label = received[0]
                if label != header and label in self.sent_acks:  # received an already received message
                    to_be_sent = (self.ID + ':' + label + ":ack:").encode()
                    self.socket.send(to_be_sent)
                    logging.info("Sent: " + label + ':ack:')
                    continue
                elif label != header and label in self.sent_messages:  # received a request for an already sent message
                    to_be_sent = (self.ID + ':' + label + ":" + self.sent_messages[label] + ':').encode()
                    self.socket.send(to_be_sent)
                    logging.info("Sent: " + label + ':' + self.sent_messages[label] + ':')
                    continue
                elif label == header:
                    dec_message = received[1]
                    to_be_sent = (self.ID + ':' + header + ":ack:").encode()
                    self.socket.send(to_be_sent)
                    self.sent_acks.append(label)
                    logging.info("Received: " + label + ":" + dec_message)
                    return dec_message
                else:
                    raise Exception
            raise ConnectionError

        except Exception as err:
            logging.error('Bob failed to receive {0}:\n{1}'.format(header, str(err)))
            sys.exit()
        except ConnectionError:
            logging.error("Bob failed to receive {0}".format(header))
            sys.exit()

    def send(self, header: str, message: str):
        """ Send method for receiver
        it listens for the request from the sender node, in case the header of the received message doesn't match it
        checks if an acknowledgment for the received header has been already sent or if the message for the requested
        header has been already sent, it sends a new acknowledgement in the first case and it sends again the message in
        the other case, it sends the expected message if the header is correct, the sent message is saved in a list
        Args:
            header (str): unique identifier
            message (str): message
        """
        try:
            for i in range(self.connection_attempts):
                received = self.recv_all()
                if not received:
                    continue
                label = received[0]
                if label != header and label in self.sent_acks:  # received an already received message
                    to_be_sent = (self.ID + ':' + label + ":ack:").encode()
                    self.socket.send(to_be_sent)
                    logging.info("Sent: " + label + ':ack:')
                    continue
                elif label != header and label in self.sent_messages:  # received a request for an already sent message
                    to_be_sent = (self.ID + ':' + label + ":" + self.sent_messages[label] + ':').encode()
                    self.socket.send(to_be_sent)
                    logging.info("Sent: " + label + ':' + self.sent_messages[label] + ':')
                    continue
                elif label == header and received[1] == 'request':
                    data = header + ':' + message + ':'
                    to_be_sent = (self.ID + ':' + data).encode()
                    self.sent_messages[header] = data
                    self.socket.send(to_be_sent)
                    logging.info('Sent: ' + header + ':' + message)
                    return
                else:
                    raise Exception
            raise ConnectionError

        except ConnectionError:
            logging.error('Bob failed to send {0}'.format(header))
            sys.exit()
        except Exception as err:
            logging.error('Bob failed to receive {0}:\n{1}'.format(header, str(err)))
            sys.exit()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module contains server's APIs and method to run it"""

import argparse
import asyncio
import logging
import sys
import time
import random
import string

import uvicorn

import QKDSimkit.core as core

from aiocache import Cache
from fastapi import FastAPI, Request, HTTPException, BackgroundTasks
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import Response
from threading import Thread

from QKDSimkit.Channel import start_channel
from QKDSimkit.core.qexceptions import cacheerror

app = FastAPI()

cache = Cache(Cache.REDIS, endpoint="localhost", port=6379, namespace="main")

origins = [
    "*"
]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


async def add_user(token: str):
    """Saves a token and its hash in memory"""
    try:
        users = {}
        hashed = core.utils.hash_token(token)
        users[hashed] = token
        await cache.set('users', users)
    except Exception:
        raise cacheerror("Failed to add user in cache")


async def add_channel(address: str):
    """Saves an address in memory"""
    try:
        await cache.set('address', address)
    except Exception:
        raise cacheerror("Failed to add channel in cache")



async def start_alice(number: int, size: int, ID: str):
    """Imports keys from an alice node, saves keys in memory

    Args:
        number (int): number of keys
        size (int): size of keys (bits)
        ID (str): identifier (hash of token)
    """
    try:
        id_keys_map = await cache.get('id_keys')
        if not id_keys_map:
            id_keys_map = {}
    except Exception:
        logging.error("Failed to retrieve key map from cache")
        sys.exit()
    try:
        address = await cache.get('address')
        if not address:
            raise Exception
    except Exception:
        logging.error("Failed to retrieve address from cache")
        sys.exit()
    key_list = []
    try:
        for i in range(number):
            res = core.alice.import_key(channel_address=address, ID=ID, size=size)
            if res == -1:
                print("Can't exchange a safe key")
            if isinstance(res, bytes):
                key_list.append(res.decode())
        id_keys_map[ID] = key_list
    except Exception as e:
        logging.error("Alice failed to exchange key " + str(e))
        return
    try:
        await cache.set('id_keys', id_keys_map)
    except Exception:
        logging.error("Failed to set key map in cache")
        sys.exit()


@app.get("/hello")
async def root(hashed: str):
    """Handshake request

    Args:
        hashed (str): hash of a pre-shared token

    Returns:
        message (str): test message for handshake
    """
    try:
        users = await cache.get('users')
        if hashed not in users.keys():
            return Response(status_code=404, content='Provided ID does not match any user')
        r = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(512))
        map_hash_r = await cache.get('proof')
        if not map_hash_r:
            map_hash_r = {}
        map_hash_r[hashed] = core.utils.hash_token(r)
        await cache.set('proof', map_hash_r)
        response = core.utils.encrypt(users[hashed], r)
        return response
    except Exception:
        logging.error("Error while validating request")


@app.get("/proof")
async def root(number: int, size: int, hashed: str, hash_proof: str, background_tasks: BackgroundTasks):
    """Chacek handshake and starts alice

    Args:
        number (int): number of keys
        size (int): size of keys (bits)
        hashed (str): identifier (hash of token)
        hash_proof (str): hash of proof message
        background_tasks: background tasks
    """
    try:
        users = await cache.get('users')
        map_hash_r = await cache.get('proof')
    except Exception:
        logging.error("Failed to retrieve data from cache")
        return Response(status_code=500, content='Internal server error')
    if hashed in users.keys() and hash_proof == map_hash_r[hashed]:
        background_tasks.add_task(start_alice, number, size, hashed)
        return "Verified!"
    return Response(status_code=404, content='Provided ID does not match any user')


@app.get("/get_key")
async def root(ID: str = 'id'):
    """Retrieves key from server

    Args:
        ID (str): hashed token
    """
    try:
        id_keys_map = await cache.get('id_keys')
    except Exception:
        logging.error("Failed to retrieve key map from cache")
    if ID not in id_keys_map.keys():
        return Response(status_code=404, content="No keys for the given ID")
    return id_keys_map[ID]


@app.middleware("http")
async def filter_get_key(request: Request, call_next):
    if request.client.host != '127.0.0.1' and request.url.path == '/get_key':
        response = Response(status_code=403, content='Forbidden')
    else:
        response = await call_next(request)
    return response


def manage_args():
    """Manages possible arguments and provides help messages"""
    parser = argparse.ArgumentParser(description='Server for Quantumacy')
    parser.add_argument('-a', '--address', default='127.0.0.1:5002', type=str,
                        help='Bind socket to this address (default: %(default)s)')
    channels = parser.add_subparsers(title='Channel type', dest='channel_type', required=True,
                                     help='Specify where the channel is (required)')
    parser_l = channels.add_parser('local', help='Run the channel on this machine')
    parser_l.add_argument('-ca', '--channel_address', default=':5000', type=str,
                          help='Bind socket to this address (default: %(default)s)')
    parser_l.add_argument('-n', '--noise', default=0.0, type=float,
                          help='Set a noise value for channel, type a float number in [0,1] (default: %(default)s)')
    parser_l.add_argument('-e', '--eve', action='store_true',
                          help='Add an eavesdropper to the channel')
    parser_e = channels.add_parser('external', help='Connect to an external channel')
    parser_e.add_argument('-ca', '--channel_address', type=str, required=True, help='Address of channel [host:port]')
    return parser.parse_args()


def start_server_and_channel(channel_address: str, noise: float, eve: bool, address: str):
    """Starts both server and channel

    Returns:
        channel_address (str): address of channel
        noise (float): ratio of noise in channel
        eve (bool): simulate an eavesdropper in channel
        address (str): where to bind this server
    """
    try:
        asyncio.run(add_user('7KHuKtJ1ZsV21DknPbcsOZIXfmH1_MnKdOIGymsQ5aA='))
        asyncio.run(add_channel(channel_address))
    except cacheerror as ce:
        logging.error(str(ce))
        sys.exit()
    _thread = Thread(target=start_channel, args=(channel_address, noise, eve))
    _thread.daemon = True
    _thread.start()
    time.sleep(0.5)
    if _thread.is_alive():
        try:
            uvicorn.run('QKDSimkit.Server:app', host=address.split(':')[0], port=int(address.split(':')[1]))
        except Exception:
            logging.error("Error on while running server")
            sys.exit()


def start_server(channel_address, address):
    """Starts server and connect to an external channel

    Args:
        channel_address (str): address of channel
        address (str): where to bind this server
    """
    try:
        asyncio.run(add_channel(channel_address))
    except cacheerror as ce:
        logging.error(str(ce))
        sys.exit()
    try:
        uvicorn.run('QKDSimkit.Server:app', host=address.split(':')[0], port=int(address.split(':')[1]))
    except Exception:
        logging.error("Error on while running server")
        sys.exit()


if __name__ == "__main__":
    args = manage_args()
    if args.channel_type == 'local':
        start_server_and_channel(args.channel_address, args.noise, args.eve, args.address)
    if args.channel_type == 'external':
        start_server(args.channel_address, args.address)



# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

import hashlib
import logging

from cryptography.fernet import Fernet


def validate(shared_key: list, other_shared_key: list) -> float:
    """It compares two keys to find differences
    Args:
        shared_key (str): first key
        other_shared_key (str): second key
    Returns:
        percent of equal elements in the two keys
    """
    if len(shared_key) > 0 and len(shared_key) == len(other_shared_key):
        i = 0
        count = 0
        while i < len(shared_key):
            if shared_key[i] == other_shared_key[i]:
                count += 1
            i += 1
        return count / len(shared_key)
    else:
        logging.error("Error")
        return -1


def encrypt(token: bytes, message: str) -> str:
    """Encrypts a message using a given key

    Args:
        token (bytes): key (32 url-safe base64-encoded bytes)
        message (str): clear text message
    Returns:
         encrypted message (str)
        """
    cipher = Fernet(token)
    enc_message = cipher.encrypt(message.encode())
    return enc_message.decode()


def decrypt(token: bytes, message: str) -> str:
    """Decrypts a message using a given key

    Args:
        token (bytes): key (32 url-safe base64-encoded bytes)
        message (str): encrypted message
    Returns:
        clear text message (str)
        """
    cipher = Fernet(token)
    dec_message = cipher.decrypt(message.encode())
    return dec_message.decode()


def hash_token(token: str):
    """hash a string
    Args:
        token (str): message
    Returns: hash (str)
    """
    h = hashlib.new('sha512_256')
    h.update(token.encode())
    return h.hexdigest()



# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module contains methods to simulate some feature of a channel"""

import logging
import random
import sys

from .models import Photon
from QKDSimkit.core.qexceptions import qnoiseerror

def eavesdropper(photon_stream) -> str:
    """Method to simulate an eavesdropper in a quantum channel

    Args:
        photon_stream (str): message containing polarizations
    Returns:
        message with new (eavesdropped) polarizations
    """
    try:
        photon_pulse = []
        new_message = '' + photon_stream[0] + ':' + photon_stream[1] + ':'
        polarization_vector = photon_stream[2].split("~")[:-1]
        for p in range(len(polarization_vector)):
            photon_pulse.append(Photon())
            photon_pulse[p].polarization = photon_pulse[p].measure(int(polarization_vector[p]))
            photon_pulse[p].bit = photon_pulse[p].set_bit_from_measurement()
            new_message += str(photon_pulse[p].polarization) + '~'
        return new_message + ':'
    except Exception as e:
        raise qnoiseerror('Eavesdropper error:\n' + str(e))


def random_errors(photon_stream, rate: float) -> str:
    """Method to simulate random errors in a quantum channel

    Args:
        photon_stream (str): message containing polarizations
        rate (float): decimal number from 0 to 1, it sets the error rate
    Returns:
        message with errors
    """
    try:
        polarization_vector = photon_stream[2].split("~")[:-1]
        new_message = '' + photon_stream[0] + ':' + photon_stream[1] + ':'
        count = 0
        for p in polarization_vector:
            if random.randint(1, 100) <= rate*100:
                polarization = random.randint(0, 3) * 45
                new_message += str(polarization) + '~'
                count += 1
            else:
                new_message += str(p) + '~'
        logging.info('Errors: ' + str(count))
        return new_message + ':'
    except Exception as e:
        raise qnoiseerror("Failed to add errors in photon stream:\n" + str(e))



# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

import logging
import socket
import sys

from .models import Photon
from .node import Node
from .qexceptions import qsocketerror, qobjecterror

logging.basicConfig(level=logging.DEBUG)


class Sender(Node):
    """Sender class, it expands Node, it contains methods to communicate a receiver node, the general idea is that this
    node dictate the communication and the receiver can just answer"""

    def __init__(self, ID, size: int):
        super().__init__(ID, size)

    def create_photon_pulse(self) -> list:
        """Create a list of photons given a size
        Returns:
             list of photons"""
        for i in range(self.photon_pulse_size):
            self.photon_pulse.append(Photon())
        self.bases = [p.basis for p in self.photon_pulse]
        return self.photon_pulse

    def send_photon_pulse(self, pulse: list):
        """Send an already created photon pulse
        it takes the polarization from each photon

        Args:
            pulse (list): photon pulse to be sent
        """
        if not isinstance(pulse, list):
            raise qobjecterror("argument must be list")
        try:
            message = ''
            for p in range(len(pulse)):
                message += str(pulse[p].polarization) + "~"
            self.send("qpulse", message)
        except socket.error:
            raise qsocketerror("not connected to any channel")

    def generate_reconciled_key(self):
        """Generate a common key between the two parties

        it checks for every photon if the chosen basis is common, if it is not common the basis is discarded
        """
        if len(self.bases) != len(self.other_bases):
            raise qobjecterror("both pulses must contain the same amount of photons")
        else:
            for i in range(len(self.bases)):
                if self.bases[i] == self.other_bases[i]:
                    self.reconciled_key.append(self.bases[i])
                else:
                    self.reconciled_key.append("")

    def send(self, header: str, message: str):
        """Sender method for sender node
        it sends a message and wait for an acknowledgment if it doesn't receive the ack in the given time
        timeout_in_seconds it may try multiple times depending on the variable connection_attempts

        Args:
            header (str): unique identifier
            message (str): string to be sent
        """
        try:
            data = header + ':' + message + ':'
            to_be_sent = (self.ID + ':' + data).encode()
            for i in range(self.connection_attempts):
                self.socket.send(to_be_sent)
                logging.info('Sent: ' + header + ':' + message)
                received = self.recv_all()
                if not received:
                    continue
                if received[0] == header and received[1] == 'ack':
                    return
            raise ConnectionError
        except Exception as err:
            logging.error('Alice failed to send {0}:\n{1}'.format(header, str(err)))
            sys.exit()
        except ConnectionError as e:
            logging.error('Alice tried to receive too many times \n' + str(e))
            sys.exit()

    def recv(self, header: str):
        """Receiver method for sender node
        It will send a request message with the given header and it will wait for the response for a time
        timeout_in_seconds it may try multiple times depending on the variable connection_attempts, every received
        message with a different header will be discarded

        Args:
            header (str): unique identifier
        """
        try:
            for i in range(self.connection_attempts):
                to_be_sent = (self.ID + ':' + header + ':request:').encode()
                self.socket.send(to_be_sent)
                received = self.recv_all()
                if not received:
                    continue
                if received[0] == header:
                    dec_message = received[1]
                    logging.info("Received: " + header + ":" + dec_message)
                    return dec_message
            raise ConnectionError
        except ConnectionError:
            raise qsocketerror("Alice tried to receive too many times ")




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module contains client's procedure to retrieve keys"""

import argparse
import http.client
import logging
import sys
import urllib.parse

import QKDSimkit.core as core
from QKDSimkit.core.qexceptions import boberror


def get_key(alice_address: str, channel_address: str, token: str, number: int, size: int):
    """Runs handshake and starts bob procedure

    Args:
        alice_address (str): address of server
        channel_address (str): address of channel
        token (str): pre shared token
        number (int): number of keys
        size (int): size of keys (bits)
    """
    try:
        hashed = core.utils.hash_token(token)
        params = urllib.parse.urlencode({'hashed': hashed})
        conn = http.client.HTTPConnection(f"{alice_address}")
        conn.request("GET", f"/hello?{params}")
        r = conn.getresponse()
        if r.status != 200:
            return r.status
        data = r.read().decode()
        proof = core.utils.decrypt(token, data)
        conn.close()
        hash_proof = core.utils.hash_token(proof)
        params = urllib.parse.urlencode({'number': number, 'size': size, 'hashed': hashed, 'hash_proof': hash_proof})
        conn.close()
        conn1 = http.client.HTTPConnection(f"{alice_address}")
        conn1.request("GET", f"/proof?{params}")
        r = conn1.getresponse()
    except Exception as e:
        logging.error("Error while connecting to server: " + str(e))
        sys.exit()
    if r.status == 200:
        try:
            key_list = []
            for n in range(number):
                res = core.bob.import_key(channel_address=channel_address, ID=hashed, size=size)
                if res == -1:
                    print("Can't exchange a safe key")
                if isinstance(res, bytes):
                    key_list.append(res.decode())
            return key_list
        except boberror as e:
            logging.error('Bob failed to exchange key: ' + str(e))
            sys.exit()
    else:
        return r.status


def manage_args():
    """Manages possible arguments and provides help messages"""

    parser = argparse.ArgumentParser(description='Client for Quantumacy')
    parser.add_argument('alice_address', type=str, help='Address of server/Alice [host:port]')
    parser.add_argument('channel_address', type=str, help='Address of channel [host:port]')
    parser.add_argument('-n', '--number', default=1, type=int, help="Number of keys (default: %(default)s)")
    parser.add_argument('-s', '--size', default=256, type=int, help="Size of keys (default: %(default)s)")
    return parser.parse_args()


def start_client(alice_address, channel_address, number, size):
    """Wrapper for get_key()
    Args:
        alice_address (str): address of server
        channel_address (str): address of channel
        number (int): number of keys
        size (int): size of keys (bits)
    """
    return get_key(alice_address, channel_address, '7KHuKtJ1ZsV21DknPbcsOZIXfmH1_MnKdOIGymsQ5aA=', number, size)


if __name__ == '__main__':
    args = manage_args()
    start_client(args.alice_address, args.channel_address, args.number, args.size)



# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module simulates the behaviour of a photon"""

import random


class Photon(object):
    """Photon, it has bit, basis and polarization and some methods to manipulate them"""
    def __init__(self):
        self.bit = self.create_random_bit()
        self.basis = self.select_random_basis()
        self.polarization = self.set_polarization()

    def create_random_bit(self) -> int:
        """Assign a random bit to the photon"""
        return random.randint(0, 1)

    def select_random_basis(self):
        """Assign a random basis to the photon"""
        return random.choice(["RL", "DG"])  # RL = Rectilinear basis; DG = Diagonal basis;

    def set_polarization(self):  # Set polarization according to settings of bits and basis.
        """Assign the right polarization to the photon according to basis and bit"""
        if (self.basis == "RL" and self.bit == 0):
            return 0
        elif (self.basis == "RL" and self.bit == 1):
            return 90
        elif (self.basis == "DG" and self.bit == 0):
            return 45
        elif (self.basis == "DG" and self.bit == 1):
            return 135

        return self.polarization

    def measure(self, polarization: int) -> int:
        """Give back a polarization according to the basis of the photon

        Args:
            polarization (int): polarization
        Returns:
            polarization
        """
        if (self.basis == "RL"):
            if polarization in [0, 90]:
                return polarization
            else:
                return random.choice([0, 90])
        elif (self.basis == "DG"):
            if polarization in [45, 135]:
                return polarization
            else:
                return random.choice([45, 135])

    def set_bit_from_measurement(self) -> int:
        """Set bits according to settings of polarization and basis.
        Returns:
            bit: return a bit for each case
        """
        if (self.basis == "RL" and self.polarization == 0):
            return 0
        elif (self.basis == "RL" and self.polarization == 90):
            return 1
        elif (self.basis == "DG" and self.polarization == 45):
            return 0
        elif (self.basis == "DG" and self.polarization == 135):
            return 1

        return self.bit




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module contains methods to operate a p2p node"""

import argparse
import asyncio
import logging
import json
import sys

import uvicorn

from aiocache import Cache
from fastapi import FastAPI, Response
from fastapi.middleware.cors import CORSMiddleware

import QKDSimkit.core as core
from QKDSimkit.core.qexceptions import aliceerror, cacheerror

logging.basicConfig(level=logging.ERROR)

cache = Cache(Cache.REDIS, endpoint="localhost", port=6379, namespace="p2p_server")

alice_app = FastAPI()
bob_app = FastAPI()


async def answer_get(number: int, size: int, ID: str, type: str):
    """Starts a node

    Args:
        number (int): number of keys
        size (int): size of keys (bits)
        ID (str): identifier of a pair of nodes
        type (str): alice or bob (sender or receiver)
    Returns:
        keys
    """
    answer = {}
    keys = []
    try:
        address = await cache.get("channel_address")
    except Exception:
        logging.error("Failed to retrieve address from cache")
        sys.exit()
    try:
        for i in range(number):
            if type == 'Alice':
                key = core.alice.import_key(channel_address=address, ID=ID, size=size)
            if type == 'Bob':
                key = core.bob.import_key(channel_address=address, ID=ID, size=size)
            if key != -1:
                keys.append({"key_ID": i, "key": key})
        answer["keys"] = keys
        return answer
    except aliceerror as e:
        logging.error(type + " failed to exchange key: " + e)
        sys.exit()


@alice_app.get("/test")
async def root(number: int = 1, size: int = 256, ID: str = 'id'):
    """http request handler for alice

        Args:
            number (int): number of keys
            size (int): size of keys (bits)
            ID (str): identifier of a pair of nodes
        Returns:
            keys
    """
    try:
        return await answer_get(number, size, ID, 'Alice')
    except Exception:
        return Response(status_code=500, content='Internal server error')


@bob_app.get("/test")
async def root(number: int = 1, size: int = 256, ID: str = 'id'):
    """http request handler for bob

        Args:
            number (int): number of keys
            size (int): size of keys (bits)
            ID (str): identifier of a pair of nodes
        Returns:
            keys
    """
    try:
        return await answer_get(number, size, ID, 'Bob')
    except Exception:
        return Response(status_code=500, content='Internal server error')

origins = [
    "*"
]

alice_app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

bob_app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


def manage_args():
    """Manages possible arguments and provides help messages"""

    parser = argparse.ArgumentParser(description='Server for Quantumacy')
    parser.add_argument('node', choices=['alice', 'bob'],
                       help='Choose how to run this node')
    parser.add_argument('-c', '--channel_address', default=':5000', type=str,
                        help='Specify the address of the channel [host:port]')
    parser.add_argument('-a', '--address', default='127.0.0.1:5003', type=str,
                        help='Bind socket to this address (default: %(default)s)')
    return parser.parse_args()


def start_p2p(node: str, address: str, channel_address: str):
    """Starts server

    Args:
        node (str): type of node [alice, bob]
        address (str): where to bind socket
    """
    asyncio.run(cache.set('channel_address', channel_address))
    uvicorn.run('QKDSimkit.p2p_servers:{}_app'.format(node), host=address.split(':')[0], port=int(address.split(':')[1]))


if __name__ == '__main__':
    args = manage_args()
    start_p2p(args.node, args.address)



# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module simulates channel's operations"""

import socket
import logging

from threading import Thread

from .channel_features import eavesdropper, random_errors
from .qexceptions import qsocketerror

logging.basicConfig(level=logging.DEBUG)


class public_channel(object):  # insecure public classical/quantum channel
    def __init__(self, address: str, noise: float, eve: bool):
        self.host = address.split(':')[0]
        self.port = address.split(':')[1]
        self.noise = noise
        self.eve = eve
        self.buffer_size = 8192
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)  # reusable socket
        self.ip_list = []  # blacklist already created connections
        self.conn_list = []
        self.full_data = []

    def initiate_channel(self, *port):
        """Start channel"""
        if len(port) > 0:
            self.port = str(port[0])
        if isinstance(self.port, str):
            self.port = int(self.port)

        try:
            self.socket.bind((self.host, self.port))
        except socket.error:
            raise qsocketerror("port {0} is occupied".format(self.port))

        self.socket.listen(1)

        logging.info("initiated the channel on {0}:{1}, waiting for clients...".format(self.host, self.port))

        while True:
            conn, addr = self.socket.accept()  # initiate new serving thread for every new connection:
            if conn not in self.conn_list:
                logging.info("{0} has connected.".format(addr))
                self.ip_list.append(addr)
                self.conn_list.append(conn)
                _thread = Thread(target=self.initiate_connection, args=(conn, addr))
                _thread.daemon = True
                _thread.start()
            else:
                logging.info(self.ip_list)

    def initiate_connection(self, conn, addr):
        """Listen for messages and broadcast them"""
        while True:
            try:
                data = conn.recv(self.buffer_size)
                message = data.decode()
                if len(message.split(':')) > 1:
                    if message.split(':')[1] == 'qpulse' and message.split(':')[2] != 'ack':
                        if self.eve:
                            message = eavesdropper(message.split(':', 2))
                        if self.noise > 0:
                            message = random_errors(message.split(':', 2), self.noise)
            except ConnectionResetError:
                break
            except ConnectionAbortedError:
                break

            if not message:
                break
            else:
                fwdMessage = "{0}:{1}".format(addr, message)
                logging.info(fwdMessage)
                for clients in self.conn_list:
                    try:
                        if clients.getpeername() != addr:
                            clients.sendall(fwdMessage.encode())
                    except OSError:
                        # old connections?
                        logging.warning("Ünknown connection, ignoring...")

        conn.close()
        self.conn_list.remove(conn)
        self.ip_list.remove(addr)

        logging.info(str(addr) + " has disconnected")
